/*
 *                           TERMS AND CONDITIONS
 *                                   FOR
 *                         OPEN SOURCE CODE LICENSE
 * 
 * Japan Registry Services Co., Ltd. ("JPRS"), a Japanese corporation
 * having its head office at Chiyoda First Bldg. East 13F 3-8-1 Nishi-Kanda,
 * Chiyoda-ku, Tokyo 101-0065, Japan, grants you the license for open source
 * code specified in EXHIBIT A the "Code" subject to the following Terms and
 * Conditions ("OSCL").
 * 
 * 1. License Grant.
 *   JPRS hereby grants you a worldwide, royalty-free, non-exclusive
 *   license, subject to third party intellectual property claims:
 *   (a) under intellectual property rights (other than patent or
 *       trademark) licensable by JPRS to use, reproduce, modify, display,
 *       perform, sublicense and distribute the Code (or portions thereof)
 *       with or without modifications, and/or as part of a derivative work;
 *       or
 *   (b) under claims of the infringement through the making, using,
 *       offering to sell and/or otherwise disposing the JPRS Revised Code
 *       (or portions thereof);
 *   (c) the licenses granted in this Section 1(a) and (b) are effective on
 *       the date JPRS first distributes the Code to you under the terms of
 *       this OSCL;
 *   (d) Notwithstanding the above stated terms, no patent license is
 *       granted:
 *       1)  for a code that you delete from the Code;
 *       2)  separate from the Code; or
 *       3)  for infringements caused by:
 *            i) modification of the Code; or
 *           ii) combination of the Code with other software or devices.
 * 
 * 2. Consents.
 *   You agree that:
 *   (a) you must include a copy of this OSCL and the notice set forth in
 *       EXHIBIT A with every copy of the Code you distribute;
 *   (b) you must include a copy of this OSCL and the notice set forth in
 *       EXHIBIT A with every copy of binary form of the Code in the
 *       documentation and/or other materials provided with the distribution;
 *   (c) you may not offer or impose any terms on any source code version
 *       that alters or restricts the applicable version of this OSCL or
 *       the recipients' rights hereunder.
 *   (d) If the terms and conditions are set forth in EXHIBIT A, you must
 *       comply with those terms and conditions.
 * 
 * 3. Proprietary Information.
 *   All trademarks, service marks, patents, copyrights, trade secrets, and
 *   other proprietary rights in or related to the Code are and will remain
 *   the exclusive property of JPRS or its licensors, whether or not
 *   specifically recognized or perfected under local law except specified
 *   in this OSCL; provided however you agree and understand that the JPRS
 *   name may not be used to endorse or promote this Code without prior
 *   written approval of JPRS.
 * 
 * 4. WARRANTY DISCLAIMER.
 *   JPRS MAKES NO REPRESENTATIONS AND WARRANTIES REGARDING THE USE OF THE
 *   CODE, NOR DOES JPRS MAKE ANY REPRESENTATIONS THAT THE CODE WILL BECOME
 *   COMMERCIALLY AVAILABLE. JPRS, ITS AFFILIATES, AND ITS SUPPLIERS DO NOT
 *   WARRANT OR REPRESENT THAT THE CODE IS FREE OF ERRORS OR THAT THE CODE
 *   IS SUITABLE FOR TRANSLATION AND/OR LOCALIZATION. THE CODE IS PROVIDED
 *   ON AN "AS IS" BASIS AND JPRS AND ITS SUPPLIERS HAVE NO OBLIGATION TO
 *   CORRECT ERRORS OR TO SUPPORT THE CODE UNDER THIS OSCL FOR ANY REASON.
 *   TO THE FULL EXTENT PERMITTED BY LAW, ALL OBLIGATIONS ARE HEREBY
 *   EXCLUDED WHETHER EXPRESS, STATUTORY OR IMPLIED UNDER LAW, COURSE OF
 *   DEALING, CUSTOM, TRADE USAGE, ORAL OR WRITTEN STATEMENT OR OTHERWISE,
 *   INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY
 *   OR FITNESS FOR A PARTICULAR PURPOSE CONCERNING THE CODE.
 * 
 * 5. NO LIABILITY.
 *   UNDER NO CIRCUMSTANCES SHALL JPRS AND/OR ITS AFFILIATES, LICENSORS, OR
 *   REPRESENTATIVES BE LIABLE FOR ANY DAMAGES INCLUDING BUT NOT LIMITED TO
 *   CONSEQUENTIAL, INDIRECT, SPECIAL, PUNITIVE OR INCIDENTAL DAMAGES,
 *   WHETHER FORESEEABLE OR UNFORESEEABLE, BASED ON YOUR CLAIMS, INCLUDING,
 *   BUT NOT LIMITED TO, CLAIMS FOR LOSS OF DATA, GOODWILL, PROFITS, USE OF
 *   MONEY, INTERRUPTION IN USE OR AVAILABILITY OF DATA, STOPPAGE, IMPLIED
 *   WARRANTY, BREACH OF CONTRACT, MISREPRESENTATION, NEGLIGENCE, STRICT
 *   LIABILITY IN TORT, OR OTHERWISE.
 * 
 * 6. Indemnification.
 *   You agree at all times to indemnify, defend and hold harmless JPRS,
 *   its subsidiaries, affiliates, employees, agents, officers and
 *   directors (collectively referred to as the "Indemnified Parties") from
 *   and against any claims, losses, judgments, actions, proceedings,
 *   damages, costs and expenses (including without limitation, reasonable
 *   legal and other fees, costs, and disbursements) incurred by any of the
 *   Indemnified Parties or in violation of any law applicable to you or
 *   your granted license to use.
 * 
 * 7. Termination.
 *   This OSCL shall be automatically terminated in the events that:
 *   (a) you fail to comply with the terms herein and fail to cure such
 *       breach within 30 days of becoming aware of the breach.
 *   (b) any part of the Code conceived and developed by a third party
 *       directly or indirectly infringes on any patent, copyright or
 *       proprietary right. In such event JPRS will announce such
 *       infringement at <http://jprs.co.jp/>.
 * 
 * 8. General.
 *   This OSCL shall be governed by, and construed and enforced in
 *   accordance with, the laws of Japan. Any litigation or arbitration
 *   between the parties shall be conducted exclusively in Tokyo, Japan
 *   except written consent of JPRS provides other venue.
 * 
 * 
 *                                EXHIBIT A
 * 
 * The original open source code of idnkit-2 is idnkit-1.0 developed and
 * conceived by Japan Network Information Center ("JPNIC"), a Japanese
 * association, Kokusai-Kougyou-Kanda Bldg 6F, 2-3-4 Uchi-Kanda,
 * Chiyoda-ku, Tokyo 101-0047, Japan, and JPRS modifies above original code
 * under following Terms and Conditions set forth by JPNIC.
 * 
 *                                  JPNIC
 * 
 * Copyright (c) 2000-2002 Japan Network Information Center.  All rights reserved.
 * 
 * By using this file, you agree to the terms and conditions set forth bellow.
 * 
 *                       LICENSE TERMS AND CONDITIONS
 * 
 * The following License Terms and Conditions apply, unless a different
 * license is obtained from Japan Network Information Center ("JPNIC"),
 * a Japanese association, Kokusai-Kougyou-Kanda Bldg 6F, 2-3-4 Uchi-Kanda,
 * Chiyoda-ku, Tokyo 101-0047, Japan.
 * 
 * 1. Use, Modification and Redistribution (including distribution of any
 *    modified or derived work) in source and/or binary forms is permitted
 *    under this License Terms and Conditions.
 * 
 * 2. Redistribution of source code must retain the copyright notices as they
 *    appear in each source code file, this License Terms and Conditions.
 * 
 * 3. Redistribution in binary form must reproduce the Copyright Notice,
 *    this License Terms and Conditions, in the documentation and/or other
 *    materials provided with the distribution. For the purposes of binary
 *    distribution the "Copyright Notice" refers to the following language:
 *    "Copyright (c) 2000-2002 Japan Network Information Center.  All rights reserved."
 * 
 * 4. The name of JPNIC may not be used to endorse or promote products
 *    derived from this Software without specific prior written approval of
 *    JPNIC.
 * 
 * 5. Disclaimer/Limitation of Liability: THIS SOFTWARE IS PROVIDED BY JPNIC
 *    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 *    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL JPNIC BE LIABLE
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *    SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *    BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *    WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *    OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *    ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * 
 *                        JPRS Public License Notice
 *                                   For
 *                                idnkit-2.
 * 
 * The contents of this file are subject to the Terms and Conditions for
 * the Open Source Code License (the "OSCL"). You may not use this file
 * except in compliance with above terms and conditions. A copy of the OSCL
 * is available at <http://jprs.co.jp/idn/>.
 * The JPRS Revised Code is idnkit-2.
 * The Initial Developer of the JPRS Revised Code is Japan Network
 * Information Center ("JPNIC"), a Japanese association,
 * Kokusai-Kougyou-Kanda Bldg 6F, 2-3-4 Uchi-Kanda, Chiyoda-ku, Tokyo
 * 101-0047, Japan.
 * "Copyright (c) 2000-2002 Japan Network Information Center.  All rights reserved."
 * "Copyright (c) 2010 Japan Registry Services Co., Ltd.  All rights reserved."
 * Contributor(s): ______________________________________.
 * 
 * If you wish to allow use of your version of this file only under the
 * above License(s) and not to allow others to use your version of this
 * file, please indicate your decision by deleting the relevant provisions
 * above and replacing them with the notice and other provisions required
 * by the above License(s). If you do not delete the relevant provisions,
 * a recipient may use your version of this file under either the above
 * License(s).
 */

#include <config.h>

#include <stddef.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <idn/assert.h>
#include <idn/debug.h>
#include <idn/logmacro.h>
#include <idn/result.h>
#include <idn/langalias.h>
#include <idn/strhash8.h>
#include <idn/util.h>

/*
 * Language alias entry.
 */
typedef struct langalias_entry *idn__langalias_entry_t;

struct langalias_entry {
	char *alias_name;		/* alias name */
	char *real_name;		/* real name */
	idn__langalias_entry_t next;	/* pointer to next entry */
};

/*
 * Langalias object.
 */
struct idn__langalias {
	idn__strhash8_t hash;                /* hash table for aliases */
	idn__langalias_entry_t entries_head; /* map entries */
	idn__langalias_entry_t entries_tail;
};

/*
 * Create a langalias object.
 */
idn_result_t
idn__langalias_create(idn__langalias_t *ctxp) {
	idn_result_t r = idn_success;
	idn__langalias_t ctx = NULL;
	idn__strhash8_t new_hash = NULL;

	assert(ctxp != NULL);
	TRACE(("idn__langalias_create()\n"));

	ctx = (idn__langalias_t)malloc(sizeof(struct idn__langalias));
	if (ctx == NULL) {
		r = idn_nomemory;
		goto ret;
	}

	ctx->hash = NULL;
	ctx->entries_head = NULL;
	ctx->entries_tail = NULL;

	r = idn__strhash8_create(&new_hash);
	if (r != idn_success)
		goto ret;
	ctx->hash = new_hash;

	*ctxp = ctx;

ret:
	if (r != idn_success) {
		free(new_hash);
		free(ctx);
	}
	TRACE(("idn__langalias_create(): %s\n", idn_result_tostring(r)));
	return (r);
}

/*
 * Release all the memory allocated for 'ctx'.
 */
void
idn__langalias_destroy(idn__langalias_t ctx) {
	idn__langalias_entry_t entry;
	idn__langalias_entry_t next;

	assert(ctx != NULL);

	TRACE(("idn__langalias_destroy()\n"));

	if (ctx->entries_head != NULL) {
		entry = ctx->entries_head;
		while (entry != NULL) {
			next = entry->next;
			free(entry->alias_name);
			free(entry->real_name);
			free(entry);
			entry = next;
		}
	}

	idn__strhash8_destroy(ctx->hash, NULL);
	free(ctx);

	TRACE(("idn__langalias_destroy(): the object is destroyed\n"));
}

/*
 * Add an alias to 'ctx'.
 */
idn_result_t
idn__langalias_add(idn__langalias_t ctx, const char *alias_name,
		   const char *real_name) {
	idn_result_t r = idn_success;
	idn__langalias_entry_t new_entry = NULL;

	assert(ctx != NULL && alias_name != NULL && real_name != NULL);

	TRACE(("idn__langalias_add(alias=\"%s\", real=\"%s\")\n",
	       idn__debug_xstring(alias_name), idn__debug_xstring(real_name)));

	new_entry = (idn__langalias_entry_t)malloc(sizeof(*new_entry));
	if (new_entry == NULL) {
		r = idn_nomemory;
		goto ret;
	}
	new_entry->alias_name = NULL;
	new_entry->real_name  = NULL;
	new_entry->next       = NULL;

	new_entry->alias_name = idn__util_strdup(alias_name);
	if (new_entry->alias_name == NULL) {
		r = idn_nomemory;
		goto ret;
	}
	idn__util_asclower(new_entry->alias_name);

	new_entry->real_name = idn__util_strdup(real_name);
	if (new_entry->real_name == NULL) {
		r = idn_nomemory;
		goto ret;
	}
	idn__util_asclower(new_entry->real_name);

	r = idn__strhash8_put(ctx->hash, new_entry->alias_name,
			      new_entry->real_name);
	if (r != idn_success)
		goto ret;

	if (ctx->entries_head == NULL) {
		ctx->entries_head = new_entry;
		ctx->entries_tail = new_entry;
	} else {
		ctx->entries_tail->next = new_entry;
		ctx->entries_tail = new_entry;
	}

ret:
	TRACE(("idn__langalias_add(): %s\n", idn_result_tostring(r)));
	if (r != idn_success) {
		if (new_entry != NULL) {
			free(new_entry->alias_name);
			free(new_entry->real_name);
			free(new_entry);
		}
	}
	return (r);
}

/*
 * Read an alias definition file and add aliases in it to 'ctx'.
 */
idn_result_t
idn__langalias_addfromfile(idn__langalias_t ctx, const char *file) {
	idn_result_t r = idn_success;
	FILE *fp = NULL;
	int line_no;
	char line[200], alias[200], real[200];

	assert(ctx != NULL && file != NULL);

	TRACE(("idn__langalias_addfromfile(file=\"%s\")\n",
	       idn__debug_xstring(file)));

	fp = fopen(file, "r");
	if (fp == NULL) {
		r = idn_nofile;
		goto ret;
	}

	for (line_no = 1; fgets(line, sizeof(line), fp) != NULL; line_no++) {
		unsigned char *p = (unsigned char *)line;

		while (isascii(*p) && isspace(*p))
			p++;
		if (*p == '#' || *p == '\n' || *p == '\0')
			continue;
		if (sscanf((char *)p, "%s %s", alias, real) == 2) {
			r = idn__langalias_add(ctx, alias, real);
			if (r != idn_success)
				break;
		} else {
			INFO(("idn__langalias_addfromfile: syntax error, "
			      "line %d\n",line_no));
			r = idn_invalid_syntax;
			break;
		}
	}

ret:
	TRACE(("idn__langalias_addfromfile(): %s\n",
	       idn_result_tostring(r)));
	return (r);
}

/*
 * Find a real name for 'alias_name'.
 */
const char *
idn__langalias_find(idn__langalias_t ctx, const char *alias_name) {
	const char *real_name;
	const char *result;

	assert(ctx != NULL && alias_name != NULL);

	TRACE(("idn__langalias_find(alias=\"%s\")\n",
	       idn__debug_xstring(alias_name)));

	real_name = idn__strhash8_get(ctx->hash, alias_name);
	result = (real_name != NULL) ? real_name : alias_name;
	
	TRACE(("idn__langalias_find(): success (real=\"%s\")\n",
	       idn__debug_xstring(result)));
	return (result);
}
