/*
 *                           TERMS AND CONDITIONS
 *                                   FOR
 *                         OPEN SOURCE CODE LICENSE
 * 
 * Japan Registry Services Co., Ltd. ("JPRS"), a Japanese corporation
 * having its head office at Chiyoda First Bldg. East 13F 3-8-1 Nishi-Kanda,
 * Chiyoda-ku, Tokyo 101-0065, Japan, grants you the license for open source
 * code specified in EXHIBIT A the "Code" subject to the following Terms and
 * Conditions ("OSCL").
 * 
 * 1. License Grant.
 *   JPRS hereby grants you a worldwide, royalty-free, non-exclusive
 *   license, subject to third party intellectual property claims:
 *   (a) under intellectual property rights (other than patent or
 *       trademark) licensable by JPRS to use, reproduce, modify, display,
 *       perform, sublicense and distribute the Code (or portions thereof)
 *       with or without modifications, and/or as part of a derivative work;
 *       or
 *   (b) under claims of the infringement through the making, using,
 *       offering to sell and/or otherwise disposing the JPRS Revised Code
 *       (or portions thereof);
 *   (c) the licenses granted in this Section 1(a) and (b) are effective on
 *       the date JPRS first distributes the Code to you under the terms of
 *       this OSCL;
 *   (d) Notwithstanding the above stated terms, no patent license is
 *       granted:
 *       1)  for a code that you delete from the Code;
 *       2)  separate from the Code; or
 *       3)  for infringements caused by:
 *            i) modification of the Code; or
 *           ii) combination of the Code with other software or devices.
 * 
 * 2. Consents.
 *   You agree that:
 *   (a) you must include a copy of this OSCL and the notice set forth in
 *       EXHIBIT A with every copy of the Code you distribute;
 *   (b) you must include a copy of this OSCL and the notice set forth in
 *       EXHIBIT A with every copy of binary form of the Code in the
 *       documentation and/or other materials provided with the distribution;
 *   (c) you may not offer or impose any terms on any source code version
 *       that alters or restricts the applicable version of this OSCL or
 *       the recipients' rights hereunder.
 *   (d) If the terms and conditions are set forth in EXHIBIT A, you must
 *       comply with those terms and conditions.
 * 
 * 3. Proprietary Information.
 *   All trademarks, service marks, patents, copyrights, trade secrets, and
 *   other proprietary rights in or related to the Code are and will remain
 *   the exclusive property of JPRS or its licensors, whether or not
 *   specifically recognized or perfected under local law except specified
 *   in this OSCL; provided however you agree and understand that the JPRS
 *   name may not be used to endorse or promote this Code without prior
 *   written approval of JPRS.
 * 
 * 4. WARRANTY DISCLAIMER.
 *   JPRS MAKES NO REPRESENTATIONS AND WARRANTIES REGARDING THE USE OF THE
 *   CODE, NOR DOES JPRS MAKE ANY REPRESENTATIONS THAT THE CODE WILL BECOME
 *   COMMERCIALLY AVAILABLE. JPRS, ITS AFFILIATES, AND ITS SUPPLIERS DO NOT
 *   WARRANT OR REPRESENT THAT THE CODE IS FREE OF ERRORS OR THAT THE CODE
 *   IS SUITABLE FOR TRANSLATION AND/OR LOCALIZATION. THE CODE IS PROVIDED
 *   ON AN "AS IS" BASIS AND JPRS AND ITS SUPPLIERS HAVE NO OBLIGATION TO
 *   CORRECT ERRORS OR TO SUPPORT THE CODE UNDER THIS OSCL FOR ANY REASON.
 *   TO THE FULL EXTENT PERMITTED BY LAW, ALL OBLIGATIONS ARE HEREBY
 *   EXCLUDED WHETHER EXPRESS, STATUTORY OR IMPLIED UNDER LAW, COURSE OF
 *   DEALING, CUSTOM, TRADE USAGE, ORAL OR WRITTEN STATEMENT OR OTHERWISE,
 *   INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY
 *   OR FITNESS FOR A PARTICULAR PURPOSE CONCERNING THE CODE.
 * 
 * 5. NO LIABILITY.
 *   UNDER NO CIRCUMSTANCES SHALL JPRS AND/OR ITS AFFILIATES, LICENSORS, OR
 *   REPRESENTATIVES BE LIABLE FOR ANY DAMAGES INCLUDING BUT NOT LIMITED TO
 *   CONSEQUENTIAL, INDIRECT, SPECIAL, PUNITIVE OR INCIDENTAL DAMAGES,
 *   WHETHER FORESEEABLE OR UNFORESEEABLE, BASED ON YOUR CLAIMS, INCLUDING,
 *   BUT NOT LIMITED TO, CLAIMS FOR LOSS OF DATA, GOODWILL, PROFITS, USE OF
 *   MONEY, INTERRUPTION IN USE OR AVAILABILITY OF DATA, STOPPAGE, IMPLIED
 *   WARRANTY, BREACH OF CONTRACT, MISREPRESENTATION, NEGLIGENCE, STRICT
 *   LIABILITY IN TORT, OR OTHERWISE.
 * 
 * 6. Indemnification.
 *   You agree at all times to indemnify, defend and hold harmless JPRS,
 *   its subsidiaries, affiliates, employees, agents, officers and
 *   directors (collectively referred to as the "Indemnified Parties") from
 *   and against any claims, losses, judgments, actions, proceedings,
 *   damages, costs and expenses (including without limitation, reasonable
 *   legal and other fees, costs, and disbursements) incurred by any of the
 *   Indemnified Parties or in violation of any law applicable to you or
 *   your granted license to use.
 * 
 * 7. Termination.
 *   This OSCL shall be automatically terminated in the events that:
 *   (a) you fail to comply with the terms herein and fail to cure such
 *       breach within 30 days of becoming aware of the breach.
 *   (b) any part of the Code conceived and developed by a third party
 *       directly or indirectly infringes on any patent, copyright or
 *       proprietary right. In such event JPRS will announce such
 *       infringement at <http://jprs.co.jp/>.
 * 
 * 8. General.
 *   This OSCL shall be governed by, and construed and enforced in
 *   accordance with, the laws of Japan. Any litigation or arbitration
 *   between the parties shall be conducted exclusively in Tokyo, Japan
 *   except written consent of JPRS provides other venue.
 * 
 * 
 *                                EXHIBIT A
 * 
 * The original open source code of idnkit-2 is idnkit-1.0 developed and
 * conceived by Japan Network Information Center ("JPNIC"), a Japanese
 * association, Kokusai-Kougyou-Kanda Bldg 6F, 2-3-4 Uchi-Kanda,
 * Chiyoda-ku, Tokyo 101-0047, Japan, and JPRS modifies above original code
 * under following Terms and Conditions set forth by JPNIC.
 * 
 *                                  JPNIC
 * 
 * Copyright (c) 2000-2002 Japan Network Information Center.  All rights reserved.
 * 
 * By using this file, you agree to the terms and conditions set forth bellow.
 * 
 *                       LICENSE TERMS AND CONDITIONS
 * 
 * The following License Terms and Conditions apply, unless a different
 * license is obtained from Japan Network Information Center ("JPNIC"),
 * a Japanese association, Kokusai-Kougyou-Kanda Bldg 6F, 2-3-4 Uchi-Kanda,
 * Chiyoda-ku, Tokyo 101-0047, Japan.
 * 
 * 1. Use, Modification and Redistribution (including distribution of any
 *    modified or derived work) in source and/or binary forms is permitted
 *    under this License Terms and Conditions.
 * 
 * 2. Redistribution of source code must retain the copyright notices as they
 *    appear in each source code file, this License Terms and Conditions.
 * 
 * 3. Redistribution in binary form must reproduce the Copyright Notice,
 *    this License Terms and Conditions, in the documentation and/or other
 *    materials provided with the distribution. For the purposes of binary
 *    distribution the "Copyright Notice" refers to the following language:
 *    "Copyright (c) 2000-2002 Japan Network Information Center.  All rights reserved."
 * 
 * 4. The name of JPNIC may not be used to endorse or promote products
 *    derived from this Software without specific prior written approval of
 *    JPNIC.
 * 
 * 5. Disclaimer/Limitation of Liability: THIS SOFTWARE IS PROVIDED BY JPNIC
 *    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 *    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL JPNIC BE LIABLE
 *    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *    SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *    BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *    WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *    OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *    ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * 
 *                        JPRS Public License Notice
 *                                   For
 *                                idnkit-2.
 * 
 * The contents of this file are subject to the Terms and Conditions for
 * the Open Source Code License (the "OSCL"). You may not use this file
 * except in compliance with above terms and conditions. A copy of the OSCL
 * is available at <http://jprs.co.jp/idn/>.
 * The JPRS Revised Code is idnkit-2.
 * The Initial Developer of the JPRS Revised Code is Japan Network
 * Information Center ("JPNIC"), a Japanese association,
 * Kokusai-Kougyou-Kanda Bldg 6F, 2-3-4 Uchi-Kanda, Chiyoda-ku, Tokyo
 * 101-0047, Japan.
 * "Copyright (c) 2000-2002 Japan Network Information Center.  All rights reserved."
 * "Copyright (c) 2010 Japan Registry Services Co., Ltd.  All rights reserved."
 * Contributor(s): ______________________________________.
 * 
 * If you wish to allow use of your version of this file only under the
 * above License(s) and not to allow others to use your version of this
 * file, please indicate your decision by deleting the relevant provisions
 * above and replacing them with the notice and other provisions required
 * by the above License(s). If you do not delete the relevant provisions,
 * a recipient may use your version of this file under either the above
 * License(s).
 */

#include <config.h>

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <idn/assert.h>
#include <idn/debug.h>
#include <idn/logmacro.h>
#include <idn/result.h>
#include <idn/utf32.h>

/*
 * Unicode surrogate pair.
 */
#define SURROGATE_HIGH(v)	(SURROGATE_H_OFF + (((v) - 0x10000) >> 10))
#define SURROGATE_LOW(v)	(SURROGATE_L_OFF + ((v) & 0x3ff))
#define SURROGATE_BASE		0x10000
#define SURROGATE_H_OFF		0xd800
#define SURROGATE_L_OFF		0xdc00
#define COMBINE_SURROGATE(h, l) \
	(SURROGATE_BASE + (((h)-SURROGATE_H_OFF)<<10) + ((l)-SURROGATE_L_OFF))

/*
 * ASCII ctype macros.
 * Note that these macros evaluate the argument multiple times.  Be careful.
 */
#define ASCII_TOUPPER(c) \
	(('a' <= (c) && (c) <= 'z') ? ((c) - 'a' + 'A') : (c))
#define ASCII_TOLOWER(c) \
	(('A' <= (c) && (c) <= 'Z') ? ((c) - 'A' + 'a') : (c))

/*
 * Convert an UTF-32 character to a multibyte UTF-16 sequence.
 */
idn_result_t
idn__utf32_toutf16(const unsigned long *utf32, unsigned short *utf16,
		   size_t tolen) {
	idn_result_t r = idn_success;
	unsigned short *utf16p = utf16;
	unsigned long v;

	TRACE(("idn__utf32_toutf16(utf32=\"%s\", tolen=%d)\n",
	       idn__debug_utf32xstring(utf32), (int)tolen));

	while (*utf32 != '\0') {
		v = *utf32++;

		if (IS_SURROGATE_LOW(v) || IS_SURROGATE_HIGH(v)) {
			WARNING(("idn__utf32_utf32toutf16: UTF-32 string "
				 "contains surrogate pair\n"));
			r = idn_invalid_encoding;
			goto ret;
		} else if (v > 0xffff) {
			/* Convert to surrogate pair */
			if (v >= 0x110000) {
				r = idn_invalid_encoding;
				goto ret;
			}
			if (tolen < 2) {
				r = idn_buffer_overflow;
				goto ret;
			}
			*utf16p++ = SURROGATE_HIGH(v);
			*utf16p++ = SURROGATE_LOW(v);
			tolen -= 2;
		} else {
			if (tolen < 1) {
				r = idn_buffer_overflow;
				goto ret;
			}
			*utf16p++ = v;
			tolen--;
		}
	}

	if (tolen < 1) {
		r = idn_buffer_overflow;
		goto ret;
	}
	*utf16p = '\0';

ret:
	if (r == idn_success) {
		TRACE(("idn__utf32_toutf16(): success (utf16=\"%s\")\n",
		       idn__debug_utf16xstring(utf16)));
	} else {
		TRACE(("idn__utf32_toutf16(): %s\n", idn_result_tostring(r)));
	}
	return (r);
}

/*
 * Convert a multibyte UTF-16 sequence to an UTF-32 character.
 */
idn_result_t
idn__utf32_fromutf16(const unsigned short *utf16, unsigned long *utf32,
		     size_t tolen) {
	idn_result_t r = idn_success;
	unsigned long *utf32p = utf32;
	unsigned short v0, v1;

	TRACE(("idn__utf32_fromutf16(utf16=\"%s\", tolen=%d)\n",
	       idn__debug_utf16xstring(utf16), (int)tolen));

	while (*utf16 != '\0') {
		v0 = *utf16;

		if (tolen < 1) {
			r = idn_buffer_overflow;
			goto ret;
		}

		if (IS_SURROGATE_HIGH(v0)) {
			v1 = *(utf16 + 1);
			if (!IS_SURROGATE_LOW(v1)) {
				WARNING(("idn__utf32_utf16toutf32: "
					 "corrupted surrogate pair\n"));
				r = idn_invalid_encoding;
				goto ret;
			}
			*utf32p++ = COMBINE_SURROGATE(v0, v1);
			tolen--;
			utf16 += 2;

		} else {
			*utf32p++ = v0;
			tolen--;
			utf16++;
			
		}
	}

	if (tolen < 1) {
		r = idn_buffer_overflow;
		goto ret;
	}
	*utf32p = '\0';

ret:
	if (r == idn_success) {
		TRACE(("idn__utf32_fromutf16(): success (utf32=\"%s\")\n",
		       idn__debug_utf32xstring(utf32)));
	} else {
		TRACE(("idn__utf32_fromutf16(): %s\n",
		       idn_result_tostring(r)));
	}
	return (r);
}

/*
 * Convert a multibyte UTF-8 sequence to an UTF-32 character.
 */
idn_result_t
idn__utf32_fromutf8(const char *utf8, unsigned long *utf32, size_t tolen) {
	idn_result_t r = idn_success;
	const unsigned char *utf8p = (const unsigned char *)utf8;
	unsigned long *utf32p = utf32;
	unsigned long v, min;
	unsigned char c;
	int width;
	int i;

	TRACE(("idn__utf32_fromutf8(utf8=\"%s\", tolen=%d)\n",
	       idn__debug_xstring(utf8), (int)tolen));

	while(*utf8p != '\0') {
		c = *utf8p++;
		if (c < 0x80) {
			v = c;
			min = 0;
			width = 1;
		} else if (c < 0xc0) {
			WARNING(("idn__utf32_fromutf8: invalid character\n"));
			r = idn_invalid_encoding;
			goto ret;
		} else if (c < 0xe0) {
			v = c & 0x1f;
			min = 0x80;
			width = 2;
		} else if (c < 0xf0) {
			v = c & 0x0f;
			min = 0x800;
			width = 3;
		} else if (c < 0xf8) {
			v = c & 0x07;
			min = 0x10000;
			width = 4;
		} else if (c < 0xfc) {
			v = c & 0x03;
			min = 0x200000;
			width = 5;
		} else if (c < 0xfe) {
			v = c & 0x01;
			min = 0x4000000;
			width = 6;
		} else {
			WARNING(("idn__utf32_fromutf8: invalid character\n"));
			r = idn_invalid_encoding;
			goto ret;
		}

		for (i = width - 1; i > 0; i--) {
			c = *utf8p++;
			if (c < 0x80 || 0xc0 <= c) {
				WARNING(("idn__utf32_fromutf8: "
					 "invalid character\n"));
				r = idn_invalid_encoding;
				goto ret;
			}
			v = (v << 6) | (c & 0x3f);
		}

	        if (v < min || v > UTF32_MAX) {
			WARNING(("idn__utf32_fromutf8: invalid character\n"));
			r = idn_invalid_encoding;
			goto ret;
		}
		if (IS_SURROGATE_LOW(v) || IS_SURROGATE_HIGH(v)) {
			WARNING(("idn__utf32_fromutf8: UTF-8 string contains "
				 "surrogate pair\n"));
			r = idn_invalid_encoding;
			goto ret;
		}
		if (tolen < 1) {
			r = idn_buffer_overflow;
			goto ret;
		}
		tolen--;
		*utf32p++ = v;
	}

	if (tolen < 1) {
		r = idn_buffer_overflow;
		goto ret;
	}
	*utf32p = '\0';

ret:
	if (r == idn_success) {
		TRACE(("idn__utf32_fromutf8(): success (utf32=\"%s\")\n",
		       idn__debug_utf32xstring(utf32)));
	} else {
		TRACE(("idn__utf32_fromutf8(): %s\n",
		       idn_result_tostring(r)));
	}
	return (r);
}

/*
 * Convert an UTF-32 character to a multibyte UTF-8 sequence.
 */
idn_result_t
idn__utf32_toutf8(const unsigned long *utf32, char *utf8, size_t tolen) {
	idn_result_t r = idn_success;
	unsigned char *utf8p = (unsigned char *)utf8;
	unsigned long v;
	int width;
	int mask;
	int offset;

	TRACE(("idn__utf32_toutf8(utf32=\"%s\", tolen=%d)\n",
	       idn__debug_utf32xstring(utf32), (int)tolen));

	while (*utf32 != '\0') {
		v = *utf32++;
		if (IS_SURROGATE_LOW(v) || IS_SURROGATE_HIGH(v)) {
			WARNING(("idn__utf32_toutf8: UTF-32 string contains "
				 "surrogate pair\n"));
			r = idn_invalid_encoding;
			goto ret;
		}
		if (v < 0x80) {
			mask = 0;
			width = 1;
		} else if (v < 0x800) {
			mask = 0xc0;
			width = 2;
		} else if (v < 0x10000) {
			mask = 0xe0;
			width = 3;
		} else if (v < 0x200000) {
			mask = 0xf0;
			width = 4;
		} else if (v < 0x4000000) {
			mask = 0xf8;
			width = 5;
		} else if (v < 0x80000000) {
			mask = 0xfc;
			width = 6;
		} else {
			WARNING(("idn__utf32_toutf8: invalid character\n"));
			r = idn_invalid_encoding;
			goto ret;
		}

		if (tolen < width) {
			r = idn_buffer_overflow;
			goto ret;
		}
		offset = 6 * (width - 1);
		*utf8p++ = (v >> offset) | mask;
		mask = 0x80;
		while (offset > 0) {
			offset -= 6;
			*utf8p++ = ((v >> offset) & 0x3f) | mask;
		}
		tolen -= width;
	}

	if (tolen < 1) {
		r = idn_buffer_overflow;
		goto ret;
	}
	*utf8p = '\0';

ret:
	if (r == idn_success) {
		TRACE(("idn__utf32_toutf8(): success (utf8=\"%s\")\n",
		       idn__debug_xstring(utf8)));
	} else {
		TRACE(("idn__utf32_toutf8(): %s\n",
		       idn_result_tostring(r)));
	}
	return (r);
}

/*
 * UTF-32 string version of strlen().
 * It returns the number of characters in a string.
 */
size_t
idn__utf32_strlen(const unsigned long *utf32) {
	size_t len;

	for (len = 0; *utf32 != '\0'; utf32++, len++)
		/* nothing to do */ ;

	return (len);
}

/*
 * UTF-32 string version of strnlen().
 * It returns the number of characters in a string, but at most 'n'.
 */
size_t
idn__utf32_strnlen(const unsigned long *utf32, size_t n) {
	size_t len;

	for (len = 0; n > 0 && *utf32 != '\0'; n--, utf32++, len++)
		/* nothing to do */ ;

	return (len);
}


/*
 * UTF-32 string version of strchr().
 */
unsigned long *
idn__utf32_strchr(const unsigned long *s, unsigned long c) {
	unsigned long *p = (unsigned long *)s;

	for (;;) {
		if (*p == c)
			return (p);
		if (*p == '\0')
			return (NULL);
		p++;
	}

	return (NULL);  /* not reached */
}

/*
 * UTF-32 string version of strrchr().
 */
unsigned long *
idn__utf32_strrchr(const unsigned long *s, unsigned long c) {
	unsigned long *p = (unsigned long *)s;
	unsigned long *result = NULL;

	for (;;) {
		if (*p == c)
			result = p;
		if (*p == '\0')
			return (result);
		p++;
	}

	return (NULL);  /* not reached */
}

/*
 * UTF-32 string version of strcpy().
 */
idn_result_t
idn__utf32_strcpy(unsigned long *to, size_t tolen, const unsigned long *from) {
	while (*from != '\0') {
		if (tolen <= 0)
			return (idn_buffer_overflow);
		*to++ = *from++;
		tolen--;
	}

	if (tolen <= 0)
		return (idn_buffer_overflow);
	*to = '\0';

	return (idn_success);
}

/*
 * UTF-32 string version of strncpy().
 */
idn_result_t
idn__utf32_strncpy(unsigned long *to, size_t tolen,
		  const unsigned long *from, size_t n) {
	while (n > 0 && *from != '\0') {
		if (tolen <= 0)
			return (idn_buffer_overflow);
		*to++ = *from++;
		tolen--;
		n--;
	}

	if (tolen <= 0)
		return (idn_buffer_overflow);
	*to = '\0';

	return (idn_success);
}

/*
 * UTF-32 string version of strcat().
 */
idn_result_t
idn__utf32_strcat(unsigned long *to, size_t tolen, const unsigned long *from) {
	while (*to != '\0') {
		if (tolen <= 0)
			return (idn_buffer_overflow);
		to++;
		tolen--;
	}

	while (*from != '\0') {
		if (tolen <= 0)
			return (idn_buffer_overflow);
		*to++ = *from++;
		tolen--;
	}

	if (tolen <= 0)
		return (idn_buffer_overflow);
	*to = '\0';

	return (idn_success);
}

/*
 * UTF-32 string version of strncat().
 */
idn_result_t
idn__utf32_strncat(unsigned long *to, size_t tolen,
		  const unsigned long *from, size_t n) {
	while (*to != '\0') {
		if (tolen <= 0)
			return (idn_buffer_overflow);
		to++;
		tolen--;
	}

	while (n > 0 && *from != '\0') {
		if (tolen <= 0)
			return (idn_buffer_overflow);
		*to++ = *from++;
		tolen--;
		n--;
	}

	if (tolen <= 0)
		return (idn_buffer_overflow);
	*to = '\0';

	return (idn_success);
}

/*
 * UTF-32 string version of strcmp().
 */
int
idn__utf32_strcmp(const unsigned long *str1, const unsigned long *str2) {
	while (*str1 != '\0') {
		if (*str1 > *str2)
			return (1);
		else if (*str1 < *str2)
			return (-1);
		str1++;
		str2++;
	}

	if (*str1 > *str2)
		return (1);
	else if (*str1 < *str2)
		return (-1);

	return (0);
}

/*
 * UTF-32 string version of strncmp().
 */
int
idn__utf32_strncmp(const unsigned long *str1, const unsigned long *str2,
		  size_t n) {
	while (n > 0 && *str1 != '\0') {
		if (*str1 > *str2)
			return (1);
		else if (*str1 < *str2)
			return (-1);
		str1++;
		str2++;
		n--;
	}

	if (n > 0) {
		if (*str1 > *str2)
			return (1);
		else if (*str1 < *str2)
			return (-1);
	}

	return (0);
}

/*
 * UTF-32 string version of strcasecmp().
 */
int
idn__utf32_strcasecmp(const unsigned long *str1, const unsigned long *str2) {
	unsigned long c1, c2;

	while (*str1 != '\0') {
		c1 = ASCII_TOLOWER(*str1);
		c2 = ASCII_TOLOWER(*str2);
		if (c1 > c2)
			return (1);
		else if (c1 < c2)
			return (-1);
		str1++;
		str2++;
	}

	c1 = ASCII_TOLOWER(*str1);
	c2 = ASCII_TOLOWER(*str2);
	if (c1 > c2)
		return (1);
	else if (c1 < c2)
		return (-1);

	return (0);
}

/*
 * UTF-32 string version of strncasecmp().
 */
int
idn__utf32_strncasecmp(const unsigned long *str1, const unsigned long *str2,
		     size_t n) {
	unsigned long c1, c2;

	while (n > 0 && *str1 != '\0') {
		c1 = ASCII_TOLOWER(*str1);
		c2 = ASCII_TOLOWER(*str2);
		if (c1 > c2)
			return (1);
		else if (c1 < c2)
			return (-1);
		str1++;
		str2++;
		n--;
	}

	if (n > 0) {
		c1 = ASCII_TOLOWER(*str1);
		c2 = ASCII_TOLOWER(*str2);
		if (c1 > c2)
			return (1);
		else if (c1 < c2)
			return (-1);
	}

	return (0);
}

/*
 * UTF-32 string version of strdup().
 */
unsigned long *
idn__utf32_strdup(const unsigned long *str) {
	size_t length = idn__utf32_strlen(str);
	unsigned long *dupstr;

	dupstr = (unsigned long *)malloc(sizeof(*str) * (length + 1));
	if (dupstr == NULL)
		return (NULL);
	memcpy(dupstr, str, sizeof(*str) * (length + 1));

	return (dupstr);
}

/*
 * UTF-32 string version of strndup().
 */
unsigned long *
idn__utf32_strndup(const unsigned long *str, size_t n) {
	size_t length = idn__utf32_strnlen(str, n);
	unsigned long *dupstr;

	dupstr = (unsigned long *)malloc(sizeof(*str) * (length + 1));
	if (dupstr == NULL)
		return (NULL);
	memcpy(dupstr, str, sizeof(*str) * (length + 1));
	*(dupstr + length) = '\0';

	return (dupstr);
}

/*
 * Convert ASCII uppercase letters ('A'..'Z') in 's' to lowercase.
 */
void
idn__utf32_asclower(unsigned long *s) {
	unsigned long *p;

	for (p = s; *p != '\0'; p++) {
		if ('A' <= *p && *p <= 'Z')
			*p += 'a' - 'A';
	}
}

